//Include the necessary libraries
#include <Wire.h>
#include <Sodaq_DS3231.h>
#include <SPI.h>
#include <SD.h>

//These constants are used for reading the battery voltage
#define ADC_AREF 3.3
#define BATVOLTPIN A6
#define BATVOLT_R1 4.7
#define BATVOLT_R2 10

//The delay between the sensor readings
#define READ_DELAY 1000

//Digital pin 11 is the MicroSD slave select pin on the Mbili
#define SD_SS_PIN 11

//The data log file
#define FILE_NAME "DataLog.txt"

//Data header
#define DATA_HEADER "Date, Time, Temperature, Voltage"

Sodaq_DS3231 RTC; //Create RTC object for DS3231 RTC come Temp Sensor 
//year, month, date, hour, min, sec and week-day(starts from 0 and goes to 6)
//writing any non-existent time-data may interfere with normal operation of the RTC.
//Take care of week-day also.*/
DateTime dt(2014, 06, 05, 11, 5, 00, 4);

void setup()
{
  //Start serial
  Serial.begin(9600);
  Serial.println(DATA_HEADER);
  
  //Initialise sensors
  setupSensors();

  //Initialise log file
  setupLogFile();
}
 
void loop()
{
   //Create the data record
  String dataRec = createDataRecord();
  
  //Save the data record to the log file
  logData(dataRec);
  
  //Echo the data to the serial connection
  Serial.println(dataRec);
 
  //Wait before taking the next reading
  delay(READ_DELAY);
}

void setupSensors()
{
  //Initialise I2C
  Wire.begin();
  
  //Initialise the DS3231 
  RTC.begin();
  
  //remember to comment this line once RTC is updated
  RTC.setDateTime(dt); //Adjust date-time as defined 'dt' above 

}

void setupLogFile()
{
  //Initialise the SD card
  if (!SD.begin(SD_SS_PIN))
  {
    Serial.println("Error: SD card failed to initialise or is missing.");
    //Hang
    while (true); 
  }
  
  //Check if the file already exists
  bool oldFile = SD.exists(FILE_NAME);  
  
  //Open the file in write mode
  File logFile = SD.open(FILE_NAME, FILE_WRITE);
  
  //Add header information if the file did not already exist
  if (!oldFile)
  {
    logFile.println(DATA_HEADER);
  }
  
  //Close the file to save it
  logFile.close();  
}

void logData(String rec)
{
  //Re-open the file
  File logFile = SD.open(FILE_NAME, FILE_WRITE);
  
  //Write the CSV data
  logFile.println(rec);
  
  //Close the file to save it
  logFile.close();  
}

String createDataRecord()
{
  //Create a String type data record in csv format
  ///Read the temperature 
  RTC.convertTemperature();            
  float temp = RTC.getTemperature();
  
  // Convert temperature voltage to string
  char buffer[14]; //make buffer large enough for 7 digits
  String temperatureS = dtostrf(temp, 7,2,buffer);
  //'7' digits including '-' negative, decimal and white space. '2' decimal places
  temperatureS.trim(); //trim whitespace, important so Ubidots will treat it as a number
  
  //Read the voltage
  int mv = getRealBatteryVoltage() * 1000.0;
 
  String data = getDateTime()+ ", ";
  data += String(temperatureS)+ "C, ";
  data += String(mv)+ "mV";
   
  return data;
}

String getDateTime()
{
  String dateTimeStr;
  
  //Create a DateTime object from the current time
  DateTime dt(RTC.makeDateTime(RTC.now().getEpoch()));

  //Convert it to a String
  dt.addToString(dateTimeStr);
  
  return dateTimeStr;  
}

float getRealBatteryVoltage()
{
  uint16_t batteryVoltage = analogRead(BATVOLTPIN);
  return (ADC_AREF / 1023.0) * (BATVOLT_R1 + BATVOLT_R2) / BATVOLT_R2 * batteryVoltage;
}
